# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from warnings import warn, simplefilter
from six import string_types
from .service import Service
from .places import Places
import descarteslabs as dl

CONST_ID_DEPRECATION_MESSAGE = (
    "Keyword arg `const_id' has been deprecated and will be removed in "
    "future versions of this software. Use the `products` "
    "argument instead. Product identifiers can be found with the "
    " products() method."
)


class Metadata(Service):
    TIMEOUT = (9.5, 120)
    """Image Metadata Service"""

    def __init__(self, url=None, token=None):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        simplefilter('always', DeprecationWarning)
        if url is None:
            url = os.environ.get("DESCARTESLABS_METADATA_URL",
                                 "https://platform-services.descarteslabs.com/metadata/v1")

        Service.__init__(self, url, token)

    def sources(self):
        """Get a list of image sources.

        Example::
            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> sources = dl.metadata.sources()
            >>> pprint(sources)
            [{'product': 'landsat:LC08:PRE:TOAR', 'sat_id': 'LANDSAT_8'}]

        """
        r = self.session.get('/sources')
        return r.json()

    def bands(self, limit=None, products=None, offset=None, wavelength=None, resolution=None, tags=None):
        """Seach for imagery data bands that you have access to.

        :param int limit: Number of results to return.
        :param int offset: Index to start at when returning results.
        :param float wavelenth: A wavelength in nm e.g 700 that the band sensor must measure.
        :param int resolution: The resolution in meters per pixel e.g 30 of the data available in this band.
        :param list(str) tags: A list of tags that the band must have in its own tag list.


        """
        params = ['limit', 'offset', 'products', 'wavelength', 'resolution', 'tags']

        args = locals()
        kwargs = {
            param: args[param]
            for param in params
            if args[param] is not None
        }

        r = self.session.post('/bands/search', json=kwargs)
        return r.json()

    def products(self, bands=None, limit=None, offset=None):
        """Search products that are available on the platform.

        :param list(str) bands: Band name(s) e.g ["red", "nir"] to filter products by.
                                Note that products must match all bands that are passed.
        :param int limit: Number of results to return.
        :param int offset: Index to start at when returning results.

        Example::
            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> products = dl.metadata.products(limit=1)
            >>> pprint(products)
                [{'Orbit': 'sun-synchronous',
                    'Spectral Bands': '6',
                    'description': 'Landsat 5 thematic mapper imagery, processed by Descartes '
                                'Labs into Top-of-atmosphere-reflectance.\\n'
                                '\\n'
                                'Landsat 5 was a low Earth orbit satellite launched on March '
                                '1, 1984 to collect imagery of the surface of Earth. A '
                                'continuation of the Landsat Program, Landsat 5 was jointly '
                                'managed by the U.S. Geological Survey (USGS) and the '
                                'National Aeronautics and Space Administration (NASA). Data '
                                "from Landsat 5 was collected and distributed from the USGS's "
                                'Center for Earth Resources Observation and Science (EROS).\\n'
                                'After 29 years in space, Landsat 5 was officially '
                                'decommissioned on June 5, 2013.[2] Near the end of its '
                                "mission, Landsat 5's use was hampered by equipment failures, "
                                'and it was largely superseded by Landsat 7 and Landsat 8 '
                                'Mission scientists anticipated the satellite will re-enter '
                                "Earth's atmosphere and disintegrate around 2034.\\n"
                                ' \\n'
                                'This information uses material from the Wikipedia article <a '
                                'href="https://en.wikipedia.org/wiki/Landsat_5">"Landsat_5"</a>, '
                                'which is released under the <a '
                                'href="https://creativecommons.org/licenses/by-sa/3.0/">Creative '
                                'Commons Attribution-Share-Alike License 3.0</a>.\\n',
                    'end_date': '11/1/2011',
                    'native_bands': ['red', 'green', 'blue', 'nir', 'swir1', 'swir2'],
                    'orbit': 'sun-synchronous',
                    'processing_level': 'TOAR',
                    'product': 'landsat:LT05:PRE:TOAR',
                    'resolution': 30,
                    'revisit': 16,
                    'sensor': 'Thematic Mapper (TM)',
                    'spectral bands': 6,
                    'start_date': '3/1/1984',
                    'swath': '185km',
                    'title': 'Landsat 5'}]
        """
        params = ['limit', 'offset', 'bands']

        args = locals()
        kwargs = {
            param: args[param]
            for param in params
            if args[param] is not None
        }

        r = self.session.post('/products/search', json=kwargs)

        return r.json()

    def available_products(self):
        """Get the list of product identifiers you have access to.

        Example::
            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> products = dl.metadata.available_products()
            >>> pprint(products)
            ['landsat:LC08:PRE:TOAR']

        """
        r = self.session.get('/products')

        return r.json()

    def summary(self, products=None, const_id=None, sat_id=None, date='acquired', part=None,
                place=None, geom=None, start_time=None, end_time=None, cloud_fraction=None,
                cloud_fraction_0=None, fill_fraction=None, pixels=None, params=None,
                dltile=None):
        """Get a summary of the results for the specified spatio-temporal query.

        :param list(str) products: Product identifier(s).
        :param list(str) const_id: Constellation identifier(s).
        :param list(str) sat_id: Satellite identifier(s).
        :param str date: The date field to use for search (e.g. `acquired`).
        :param str part: Part of the date to aggregate over (e.g. `day`).
        :param str place: A slug identifier to be used as a region of interest.
        :param str geom: A GeoJSON or WKT region of interest.
        :param str start_time: Desired starting date and time (inclusive).
        :param str end_time: Desired ending date and time (inclusive).
        :param float cloud_fraction: Maximum cloud fraction, calculated by data provider.
        :param float cloud_fraction_0: Maximum cloud fraction, calculated by cloud mask pixels.
        :param float fill_fraction: Minimum scene fill fraction, calculated as valid/total pixels.
        :param bool pixels: Whether to include pixel counts in summary calculations.
        :param str params: JSON of additional query parameters.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.

        Example usage::

            >>> import descarteslabs as dl
            >>> from pprint import  pprint
            >>> pprint(dl.metadata.summary(place='north-america_united-states_iowa', \
                    products=['landsat:LC08:PRE:TOAR'], start_time='2016-07-06', \
                    end_time='2016-07-07', part='hour', pixels=True))
            {'bytes': 93298309,
             'count': 1,
             'items': [{'bytes': 93298309,
                        'count': 1,
                        'date': '2016-07-06T16:00:00',
                        'pixels': 250508160,
                        'timestamp': 1467820800}],
             'pixels': 250508160,
             'products': ['landsat:LC08:PRE:TOAR']}
        """
        if place:
            places = Places()
            places.auth = self.auth
            shape = places.shape(place, geom='low')
            geom = json.dumps(shape['geometry'])

        if dltile is not None:
            if isinstance(dltile, string_types):
                dltile = dl.raster.dltile(dltile)
            if isinstance(dltile, dict):
                geom = dltile['geometry']

        if isinstance(geom, dict):
            geom = json.dumps(geom)

        kwargs = {}

        if sat_id:
            if isinstance(sat_id, string_types):
                sat_id = [sat_id]

            kwargs['sat_id'] = sat_id

        if products:
            if isinstance(products, string_types):
                products = [products]

            kwargs['products'] = products

        if const_id:
            warn(CONST_ID_DEPRECATION_MESSAGE, DeprecationWarning)
            if isinstance(const_id, string_types):
                const_id = [const_id]

            kwargs['const_id'] = const_id

        if date:
            kwargs['date'] = date

        if part:
            kwargs['part'] = part

        if geom:
            kwargs['geom'] = geom

        if start_time:
            kwargs['start_time'] = start_time

        if end_time:
            kwargs['end_time'] = end_time

        if cloud_fraction is not None:
            kwargs['cloud_fraction'] = cloud_fraction

        if cloud_fraction_0 is not None:
            kwargs['cloud_fraction_0'] = cloud_fraction_0

        if fill_fraction is not None:
            kwargs['fill_fraction'] = fill_fraction

        if pixels:
            kwargs['pixels'] = pixels

        if params:
            kwargs['params'] = json.dumps(params)

        r = self.session.post('/summary', json=kwargs)
        return r.json()

    def search(self, products=None, const_id=None, sat_id=None, date='acquired', place=None,
               geom=None, start_time=None, end_time=None, cloud_fraction=None,
               cloud_fraction_0=None, fill_fraction=None, params=None,
               limit=100, offset=0, fields=None, dltile=None, sort_field=None, sort_order="asc"):
        """Search metadata given a spatio-temporal query. All parameters are
        optional. Results are paged using limit and offset. Please note offset
        plus limit cannot exceed 10000.

        :param list(str) products: Product Identifier(s).
        :param list(str) const_id: Constellation Identifier(s).
        :param list(str) sat_id: Satellite identifier(s).
        :param str date: The date field to use for search (e.g. `acquired`).
        :param str place: A slug identifier to be used as a region of interest.
        :param str geom: A GeoJSON or WKT region of interest.
        :param str start_time: Desired starting date and time (inclusive).
        :param str end_time: Desired ending date and time (inclusive).
        :param float cloud_fraction: Maximum cloud fraction, calculated by data provider.
        :param float cloud_fraction_0: Maximum cloud fraction, calculated by cloud mask pixels.
        :param float fill_fraction: Minimum scene fill fraction, calculated as valid/total pixels.
        :param str params: JSON of additional query parameters.
        :param int limit: Number of items to return. (max of 10000)
        :param int offset: Number of items to skip.
        :param list(str) fields: Properties to return.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param str sort_field: Property to sort on.
        :param str sort_order: Order of sort.

        return: GeoJSON ``FeatureCollection``

        Example::

            >>> import descarteslabs as dl
            >>> scenes = dl.metadata.search(place='north-america_united-states_iowa', \
                                         products=['landsat:LC08:PRE:TOAR'], \
                                         start_time='2016-07-01', \
                                         end_time='2016-07-31T23:59:59')
            >>> len(scenes['features'])
            1
        """
        if place:
            places = Places()
            places.auth = self.auth
            shape = places.shape(place, geom='low')
            geom = json.dumps(shape['geometry'])

        if dltile is not None:
            if isinstance(dltile, string_types):
                dltile = dl.raster.dltile(dltile)
            if isinstance(dltile, dict):
                geom = dltile['geometry']

        if isinstance(geom, dict):
            geom = json.dumps(geom)

        kwargs = {'date': date, 'limit': limit, 'offset': offset}

        if sat_id:

            if isinstance(sat_id, string_types):
                sat_id = [sat_id]

            kwargs['sat_id'] = sat_id

        if products:

            if isinstance(products, string_types):
                products = [products]

            kwargs['products'] = products

        if const_id:
            warn(CONST_ID_DEPRECATION_MESSAGE, DeprecationWarning)

            if isinstance(const_id, string_types):
                const_id = [const_id]

            kwargs['const_id'] = const_id

        if geom:
            kwargs['geom'] = geom

        if start_time:
            kwargs['start_time'] = start_time

        if end_time:
            kwargs['end_time'] = end_time

        if cloud_fraction is not None:
            kwargs['cloud_fraction'] = cloud_fraction

        if cloud_fraction_0 is not None:
            kwargs['cloud_fraction_0'] = cloud_fraction_0

        if fill_fraction is not None:
            kwargs['fill_fraction'] = fill_fraction

        if params:
            kwargs['params'] = json.dumps(params)

        if fields is not None:
            kwargs['fields'] = fields

        if sort_field is not None:
            kwargs['sort_field'] = sort_field

            if sort_order is not None:
                kwargs['sort_order'] = sort_order

        r = self.session.post('/search', json=kwargs)

        return {'type': 'FeatureCollection', "features": r.json()}

    def ids(self, products=None, const_id=None, sat_id=None, date='acquired', place=None,
            geom=None, start_time=None, end_time=None, cloud_fraction=None,
            cloud_fraction_0=None, fill_fraction=None, params=None, limit=100,
            offset=0, dltile=None, sort_field=None, sort_order='asc'):
        """Search metadata given a spatio-temporal query. All parameters are
        optional. Results are paged using limit/offset.

        :param list(str) products: Products identifier(s).
        :param list(str) const_id: Constellation identifier(s).
        :param list(str) sat_id: Satellite identifier(s).
        :param str date: The date field to use for search (e.g. `acquired`).
        :param str place: A slug identifier to be used as a region of interest.
        :param str geom: A GeoJSON or WKT region of interest.
        :param str start_time: Desired starting date and time (inclusive).
        :param str end_time: Desired ending date and time (inclusive).
        :param float cloud_fraction: Maximum cloud fraction, calculated by data provider.
        :param float cloud_fraction_0: Maximum cloud fraction, calculated by cloud mask pixels.
        :param float fill_fraction: Minimum scene fill fraction, calculated as valid/total pixels.
        :param str params: JSON of additional query parameters.
        :param int limit: Number of items to return.
        :param int offset: Number of items to skip.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param str sort_field: Property to sort on.
        :param str sort_order: Order of sort.

        :return: List of image identifiers.

        Example::

            >>> import descarteslabs as dl
            >>> ids = dl.metadata.ids(place='north-america_united-states_iowa', \
                                 products=['landsat:LC08:PRE:TOAR'], \
                                 start_time='2016-07-01', \
                                 end_time='2016-07-31T23:59:59')
            >>> len(ids)
            1

            >>> ids
            ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1']

        """
        result = self.search(sat_id=sat_id, products=products, const_id=const_id, date=date,
                             place=place, geom=geom, start_time=start_time,
                             end_time=end_time, cloud_fraction=cloud_fraction,
                             cloud_fraction_0=cloud_fraction_0, fill_fraction=fill_fraction,
                             params=params, limit=limit, offset=offset, fields=[], dltile=dltile,
                             sort_field=sort_field, sort_order=sort_order)

        return [feature['id'] for feature in result['features']]

    def keys(self, products=None, const_id=None, sat_id=None, date='acquired', place=None,
             geom=None, start_time=None, end_time=None, cloud_fraction=None,
             cloud_fraction_0=None, fill_fraction=None, params=None, limit=100,
             offset=0, dltile=None, sort_field=None, sort_order='asc'):
        """Search metadata given a spatio-temporal query. All parameters are
        optional. Results are paged using limit/offset.

        :param list(str) products: Products identifier(s).
        :param list(str) const_id: Constellation identifier(s).
        :param list(str) sat_id: Satellite identifier(s).
        :param str date: The date field to use for search (e.g. `acquired`).
        :param str place: A slug identifier to be used as a region of interest.
        :param str geom: A GeoJSON or WKT region of interest.
        :param str start_time: Desired starting date and time (inclusive).
        :param str end_time: Desired ending date and time (inclusive).
        :param float cloud_fraction: Maximum cloud fraction, calculated by data provider.
        :param float cloud_fraction_0: Maximum cloud fraction, calculated by cloud mask pixels.
        :param float fill_fraction: Minimum scene fill fraction, calculated as valid/total pixels.
        :param str params: JSON of additional query parameters.
        :param int limit: Number of items to return.
        :param int offset: Number of items to skip.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param str sort_field: Property to sort on.
        :param str sort_order: Order of sort.

        :return: List of image identifiers.

        Example::

            >>> import descarteslabs as dl
            >>> keys = dl.metadata.keys(place='north-america_united-states_iowa', \
                                 products=['landsat:LC08:PRE:TOAR'], \
                                 start_time='2016-07-01', \
                                 end_time='2016-07-31T23:59:59')
            >>> len(keys)
            1

            >>> keys
            ['meta_LC80270312016188_v1']

        """
        result = self.search(sat_id=sat_id, products=products, const_id=const_id, date=date,
                             place=place, geom=geom, start_time=start_time,
                             end_time=end_time, cloud_fraction=cloud_fraction,
                             cloud_fraction_0=cloud_fraction_0, fill_fraction=fill_fraction,
                             params=params, limit=limit, offset=offset, fields=["key"], dltile=dltile,
                             sort_field=sort_field, sort_order=sort_order)

        return [feature['key'] for feature in result['features']]

    def features(self, products=None, const_id=None, sat_id=None, date='acquired', place=None,
                 geom=None, start_time=None, end_time=None, cloud_fraction=None,
                 cloud_fraction_0=None, fill_fraction=None, params=None,
                 limit=100, dltile=None, sort_field=None, sort_order='asc'):

        """Generator that combines summary and search to page through results.

        :param int limit: Number of features to fetch per request.

        :return: Generator of GeoJSON ``Feature`` objects.
        """
        summary = self.summary(sat_id=sat_id, products=products, const_id=None, date=date,
                               place=place, geom=geom, start_time=start_time,
                               end_time=end_time, cloud_fraction=cloud_fraction,
                               cloud_fraction_0=cloud_fraction_0, fill_fraction=fill_fraction,
                               params=params, dltile=dltile)

        offset = 0
        count = summary['count']

        while offset < count:
            features = self.search(sat_id=sat_id, products=products, const_id=None,
                                   date=date, place=place, geom=geom,
                                   start_time=start_time, end_time=end_time,
                                   cloud_fraction=cloud_fraction,
                                   cloud_fraction_0=cloud_fraction_0,
                                   fill_fraction=fill_fraction, params=params,
                                   limit=limit, offset=offset,
                                   dltile=dltile, sort_field=sort_field,
                                   sort_order=sort_order)

            offset = limit + offset

            for feature in features['features']:
                yield feature

    def get(self, key):
        """Get metadata of a single image.

        :param str key: Image identifier.

        Example::

            >>> import descarteslabs as dl
            >>> meta = dl.metadata.get('meta_LC80270312016188_v1')
            >>> keys = list(meta.keys())
            >>> keys.sort()
            >>> keys
            ['acquired', 'area', 'bits_per_pixel', 'bright_fraction', 'bucket', 'cloud_fraction',
             'cloud_fraction_0', 'cs_code', 'descartes_version', 'file_md5s', 'file_sizes', 'files',
             'fill_fraction', 'geolocation_accuracy', 'geometry', 'geotrans', 'id', 'identifier', 'key',
             'processed', 'product', 'projcs', 'published', 'raster_size', 'reflectance_scale', 'roll_angle',
             'sat_id', 'solar_azimuth_angle', 'solar_elevation_angle', 'sw_version', 'terrain_correction',
             'tile_id']
        """
        r = self.session.get('/get/%s' % key)
        return r.json()

    def get_product(self, product_id):
        """Get information about a single product.

        :param str product_id: Product Identifier.

        Example::

            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> product = dl.metadata.get_product('landsat:LC08:PRE:TOAR')
            >>> pprint(product)
            {'Orbit': 'sun-synchronous',
                'Spectral Bands': '7',
                'description': 'Landsat 8 Operational Land Imager imagery, processed by '
                                'Descartes Labs into Top-of-atmosphere-reflectance. Red, '
                                'green, and blue bands are pansharpened to 15 meter resolution '
                                'using the panchromatic band.\\n'
                                ' \\n'
                                'Landsat 8 is an American Earth observation satellite launched '
                                'on February 11, 2013. It is the eighth satellite in the '
                                'Landsat program; the seventh to reach orbit successfully. '
                                'Originally called the Landsat Data Continuity Mission (LDCM), '
                                'it is a collaboration between NASA and the United States '
                                'Geological Survey(USGS). NASA Goddard Space Flight Center in '
                                'Greenbelt, Maryland, provided development, mission systems '
                                'engineering, and acquisition of the launch vehicle while the '
                                'USGS provided for development of the ground systems and will '
                                'conduct on-going mission operations.\\n'
                                'The satellite was built by Orbital Sciences Corporation, who '
                                'served as prime contractor for the mission.[3] The '
                                "spacecraft's instruments were constructed by Ball Aerospace "
                                "and NASA's Goddard Space Flight Center,[4] and its launch was "
                                'contracted to United Launch Alliance.[5] During the first 108 '
                                'days in orbit, LDCM underwent checkout and verification by '
                                'NASA and on 30 May 2013 operations were transferred from NASA '
                                'to the USGS when LDCM was officially renamed to Landsat '
                                '8.[6]\\n'
                                'This information uses material from the Wikipedia article <a '
                                'href="https://en.wikipedia.org/wiki/Landsat_8">"Landsat_8"</a>, '
                                'which is released under the <a '
                                'href="https://creativecommons.org/licenses/by-sa/3.0/">Creative '
                                'Commons Attribution-Share-Alike License 3.0</a>.\\n',
                'end_date': 'present',
                'id': 'landsat:LC08:PRE:TOAR',
                'native_bands': ['coastal-aerosol',
                                'red',
                                'green',
                                'blue',
                                'nir',
                                'swir1',
                                'swir2',
                                'tirs1',
                                'cirrus',
                                'qa_water',
                                'qa_snow',
                                'qa_cloud',
                                'qa_cirrus'],
                'notes': 'Red, green, and blue bands have been pansharpened to 15m resolution',
                'orbit': 'sun-synchronous',
                'processing_level': 'TOAR',
                'product': 'landsat:LC08:PRE:TOAR',
                'resolution': 15,
                'revisit': 16,
                'sensor': 'Operational Land Imager (OLI), Thermal Infrared Sensor (TIRS)',
                'spectral bands': 7,
                'start_date': '2/1/2013',
                'swath': '185km',
                'title': 'Landsat 8'}

        """
        r = self.session.get('/products/%s' % product_id)
        return r.json()

    def get_band(self, band_id):
        """Get information about a single product.

        :param str band_id: Band Identifier.

        """
        r = self.session.get('/bands/%s' % band_id)
        return r.json()



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import shutil
import unittest
import json

import descarteslabs as dl
from descarteslabs.addons import numpy as np


class TestRaster(unittest.TestCase):
    raster = None
    places = None

    @classmethod
    def setUpClass(cls):
        cls.raster = dl.raster
        cls.places = dl.places

    def test_raster(self):
        # test with product key
        r = self.raster.raster(
            inputs=['meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
        )
        self.assertTrue("metadata" in r)
        self.assertTrue("files" in r)
        self.assertTrue("meta_LC80270312016188_v1_red-green-blue-alpha.tif" in r['files'])

        # test with product id
        r = self.raster.raster(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
        )
        self.assertTrue("metadata" in r)
        self.assertTrue("files" in r)
        self.assertTrue("landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1_red-green-blue-alpha.tif" in r['files'])

    def test_raster_save(self):
        tmpdir = tempfile.mkdtemp()
        try:
            response = self.raster.raster(
                inputs=['meta_LC80270312016188_v1'],
                bands=['red', 'green', 'blue', 'alpha'],
                resolution=960, save=True,
                outfile_basename="{}/my-raster".format(tmpdir)
            )
            with open("{}/my-raster.tif".format(tmpdir), "rb") as f:
                f.seek(0, os.SEEK_END)
                length = f.tell()
            self.assertEqual(
                length,
                len(response['files']["{}/my-raster.tif".format(tmpdir)])
            )
        finally:
            shutil.rmtree(tmpdir)

    def test_ndarray(self):
        try:
            data, metadata = self.raster.ndarray(
                    inputs=['meta_LC80270312016188_v1'],
                    bands=['red', 'green', 'blue', 'alpha'],
                    resolution=960,
                    )
            self.assertEqual(data.shape, (249, 245, 4))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 4)
        except ImportError:
            pass

    def test_ndarray_single_band(self):
        try:
            data, metadata = self.raster.ndarray(
                    inputs=['meta_LC80270312016188_v1'],
                    bands=['red'],
                    resolution=960,
                    )
            self.assertEqual(data.shape, (249, 245))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 1)
        except ImportError:
            pass

    def test_cutline_dict(self):
        shape = {"geometry":
                 {"type": "Polygon",
                  "coordinates": [[[-95.2989209, 42.7999878], [-93.1167728, 42.3858464],
                                   [-93.7138666, 40.703737], [-95.8364984, 41.1150618],
                                   [-95.2989209, 42.7999878]]]
                  }
                 }
        try:
            data, metadata = self.raster.ndarray(
                    inputs=['meta_LC80270312016188_v1'],
                    bands=['red'],
                    resolution=960,
                    cutline=shape,
                    )
            self.assertEqual(data.shape, (245, 238))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 1)
        except ImportError:
            pass

    def test_cutline_str(self):
        shape = {"geometry":
                 {"type": "Polygon",
                  "coordinates": [[[-95.2989209, 42.7999878], [-93.1167728, 42.3858464],
                                   [-93.7138666, 40.703737], [-95.8364984, 41.1150618],
                                   [-95.2989209, 42.7999878]]]}}
        try:
            data, metadata = self.raster.ndarray(
                    inputs=['meta_LC80270312016188_v1'],
                    bands=['red'],
                    resolution=960,
                    cutline=json.dumps(shape),
                    )
            self.assertEqual(data.shape, (245, 238))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 1)
        except ImportError:
            pass

    def test_thumbnail(self):
        r = self.raster.raster(
            inputs=['meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            dimensions=[256, 256],
            scales=[[0, 4000]] * 4,
            output_format='PNG',
            data_type='Byte',
        )
        self.assertTrue("metadata" in r)
        self.assertTrue("files" in r)
        self.assertIsNotNone(r['files']['meta_LC80270312016188_v1_red-green-blue-alpha.png'])

    def test_get_bands_by_key(self):
        r = self.raster.get_bands_by_key('meta_LC80270312016188_v1')
        for band in ['red', 'green', 'blue', 'alpha', 'swir1', 'swir2', 'ndvi',
                     'ndwi', 'evi', 'cirrus']:
            self.assertTrue(band in r)

    def test_landsat8_bands(self):
        r = self.raster.get_bands_by_constellation('L8')
        for band in ['red', 'green', 'blue', 'alpha', 'swir1', 'swir2', 'ndvi',
                     'ndwi', 'evi', 'cirrus']:
            self.assertTrue(band in r)

    def test_dltiles_from_place(self):
        iowa = self.places.shape('north-america_united-states_iowa', geom='low')
        iowa_geom = iowa['geometry']
        dltiles_feature_collection = self.raster.dltiles_from_shape(30.0, 2048, 16, iowa_geom)
        self.assertEqual(len(dltiles_feature_collection['features']), 58)

    def test_dltiles_from_latlon(self):
        dltile_feature = self.raster.dltile_from_latlon(45.0, -90.0, 30.0, 2048, 16)
        self.assertEqual(dltile_feature['properties']['key'], "2048:16:30.0:16:-4:81")

    def test_dltile(self):
        dltile_feature = self.raster.dltile("2048:16:30.0:16:-4:81")
        self.assertEqual(dltile_feature['properties']['key'], "2048:16:30.0:16:-4:81")

    def test_raster_dltile(self):
        dltile_feature = self.raster.dltile_from_latlon(41.0, -94.0, 30.0, 256, 16)
        arr, meta = self.raster.ndarray(
            inputs=['meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            dltile=dltile_feature['properties']['key'],
        )
        self.assertEqual(arr.shape[0], 256 + 2 * 16)
        self.assertEqual(arr.shape[1], 256 + 2 * 16)
        self.assertEqual(arr.shape[2], 4)

    def test_raster_dltile_dict(self):
        dltile_feature = self.raster.dltile_from_latlon(41.0, -94.0, 30.0, 256, 16)
        arr, meta = self.raster.ndarray(
            inputs=['meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            dltile=dltile_feature,
        )
        self.assertEqual(arr.shape[0], 256 + 2 * 16)
        self.assertEqual(arr.shape[1], 256 + 2 * 16)
        self.assertEqual(arr.shape[2], 4)

    def test_dlkeys_from_place(self):
        iowa = self.places.shape('north-america_united-states_iowa', geom='low')
        iowa_geom = iowa['geometry']
        dlkeys_geojson = self.raster.dlkeys_from_shape(30.0, 2048, 16, iowa_geom)
        self.assertEqual(len(dlkeys_geojson['features']), 58)

    def test_dlkeys_from_latlon(self):
        dlkey_geojson = self.raster.dlkey_from_latlon(45.0, -90.0, 30.0, 2048, 16)
        self.assertEqual(dlkey_geojson['properties']['key'], "2048:16:30.0:16:-4:81")

    def test_dlkey(self):
        dlkey_geojson = self.raster.dlkey("2048:16:30.0:16:-4:81")
        self.assertEqual(dlkey_geojson['properties']['key'], "2048:16:30.0:16:-4:81")



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
import os
from functools import partial
from cachetools import TTLCache, cachedmethod
from cachetools.keys import hashkey
from .service import Service


class Places(Service):
    TIMEOUT = (9.5, 120)
    """Places and statistics service https://iam.descarteslabs.com/service/waldo"""

    def __init__(self, url=None, token=None, maxsize=10, ttl=600):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        if url is None:
            url = os.environ.get("DESCARTESLABS_PLACES_URL", "https://platform-services.descarteslabs.com/waldo/v1")

        Service.__init__(self, url, token)
        self.cache = TTLCache(maxsize, ttl)

    def placetypes(self):
        """Get a list of place types.

        Example::
            >>> import descarteslabs as dl
            >>> dl.places.placetypes()
            ['continent', 'country', 'dependency', 'macroregion', 'region',
                'district', 'mesoregion', 'microregion', 'county', 'locality']
        """
        r = self.session.get('/placetypes')
        return r.json()

    @cachedmethod(operator.attrgetter('cache'), key=partial(hashkey, 'find'))
    def find(self, path, **kwargs):
        """Find candidate slugs based on full or partial path.

        :param str path: Candidate underscore-separated slug.
        :param placetype: Optional place type for filtering.

        Example::

            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> results = dl.places.find('morocco')
            >>> _ = results[0].pop('bbox')
            >>> pprint(results)
            [{'id': 85632693,
              'name': 'Morocco',
              'path': 'continent:africa_country:morocco',
              'placetype': 'country',
              'slug': 'africa_morocco'}]
        """
        r = self.session.get('/find/%s' % path, params=kwargs)
        return r.json()

    @cachedmethod(operator.attrgetter('cache'), key=partial(hashkey, 'shape'))
    def shape(self, slug, output='geojson', geom='low'):
        """Get the geometry for a specific slug

        :param slug: Slug identifier.
        :param str output: Desired geometry format (`GeoJSON`).
        :param str geom: Desired resolution for the geometry (`low`, `medium`, `high`).

        :return: GeoJSON ``Feature``

        Example::
            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> kansas = dl.places.shape('north-america_united-states_kansas')
            >>> kansas['bbox']
            [-102.051744, 36.993016, -94.588658, 40.003078]

            >>> kansas['geometry']['type']
            'Polygon'

            >>> pprint(kansas['properties'])
            {'name': 'Kansas',
             'parent_id': 85633793,
             'path': 'continent:north-america_country:united-states_region:kansas',
             'placetype': 'region',
             'slug': 'north-america_united-states_kansas'}

        """
        r = self.session.get('/shape/%s.%s' % (slug, output), params={'geom': geom})
        return r.json()

    @cachedmethod(operator.attrgetter('cache'), key=partial(hashkey, 'prefix'))
    def prefix(self, slug, output='geojson', placetype=None, geom='low'):
        """Get all the places that start with a prefix

        :param str slug: Slug identifier.
        :param str output: Desired geometry format (`GeoJSON`, `TopoJSON`).
        :param str placetype: Restrict results to a particular place type.
        :param str geom: Desired resolution for the geometry (`low`, `medium`, `high`).

        :return: GeoJSON or TopoJSON ``FeatureCollection``

        Example::
            >>> import descarteslabs as dl
            >>> il_counties = dl.places.prefix('north-america_united-states_illinois', placetype='county')
            >>> len(il_counties['features'])
            102

        """
        params = {}
        if placetype:
            params['placetype'] = placetype
        params['geom'] = geom
        r = self.session.get('/prefix/%s.%s' % (slug, output), params=params)

        return r.json()



#!/usr/bin/env python
# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import argparse
import json
import descarteslabs as dl


def scales(s):
    try:
        if s.count(",") == 1:
            mi, ma = map(float, s.split(","))
            return [mi, ma]
        elif s.count(",") == 3:
            smi, sma, dmi, dma = map(float, s.split(","))
            return [smi, sma, dmi, dma]
        else:
            raise
    except:
        raise argparse.ArgumentTypeError(
            "Scales must be mi,ma or smi,sma,dmi,dma"
        )


def raster_handler(args):
    params = {
        'inputs': args.inputs,
        'bands': args.bands,
        'scales': args.scales,
        'resolution': args.resolution,
        'data_type': args.data_type,
        'output_format': args.output_format,
        'srs': args.srs,
        'place': args.place,
        'save': True,
        'outfile_basename': args.outfile_basename,
    }
    response = dl.raster.raster(**params)
    print(json.dumps(response['metadata'], indent=2))



#!/usr/bin/env python
# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import argparse

from descarteslabs.scripts.parser.auth import auth_handler
from descarteslabs.scripts.parser.metadata import metadata_handler
from descarteslabs.scripts.parser.raster import scales, raster_handler
from descarteslabs.scripts.parser.places import places_handler

parser = argparse.ArgumentParser()
subparsers = parser.add_subparsers(dest='group', )

# Auth Group
auth_parser = subparsers.add_parser('auth')
auth_parser.add_argument(
    'command',
    choices=['login', 'token', 'name', 'groups', 'payload', 'env', 'version'],
    help='The action to take (e.g. login, etc.)'
)

# Metadata Group
metadata_parser = subparsers.add_parser('metadata')
metadata_parser.add_argument('command', choices=['sources', 'summary', 'search', 'keys', 'get'],
                             help='The action to take')
metadata_parser.add_argument('argument', nargs='?')
metadata_parser.add_argument('-url', help='The url of the service')
metadata_parser.add_argument('-place', help='A slug for a named place')
metadata_parser.add_argument('-const_id', default=None, help='Constellation ID', nargs='+')
metadata_parser.add_argument('-start_time', help='Start of valid date/time range (inclusive)')
metadata_parser.add_argument('-end_time', help='End of valid date/time range (inclusive)')
metadata_parser.add_argument('-geom', help='Region of interest as GeoJSON or WKT')
metadata_parser.add_argument('-params',
                             help='JSON String of additional key/value pairs for searching properties: tile_id, cloud_fraction, etc.')  # NOQA
metadata_parser.add_argument('-limit', default=100, help='Number of items to return (default 100)', type=int)
metadata_parser.add_argument('-offset', help='Number of items to skip (default 0)')
metadata_parser.add_argument('-bbox', help='Whether or not to use a bounding box filter (default: false)',
                             action='store_true')

# Places Group
places_parser = subparsers.add_parser('places')
places_parser.add_argument('command', choices=['find', 'shape', 'prefix', 'placetypes'],
                           help='The action to take.')
places_parser.add_argument('argument', nargs='?')

places_parser.add_argument('-url', help='The url of the service')
places_parser.add_argument('-placetype', default=None, help='The placetype of the response')
places_parser.add_argument('-geom', default='low',
                           choices=['low', 'medium', 'high', "None"],
                           help='Resolution of shape')
places_parser.add_argument("-output", default='geojson')

# Raster Group
raster_parser = subparsers.add_parser('raster')
raster_parser.add_argument("inputs", type=str, nargs="+")
raster_parser.add_argument("-bands", nargs="+", default=None, type=str)
raster_parser.add_argument("-scales", default=None, type=scales, nargs="+")
raster_parser.add_argument("-resolution", default=240, type=float)
raster_parser.add_argument("-output_format", default='GTiff', type=str)
raster_parser.add_argument("-data_type", default='UInt16', type=str)
raster_parser.add_argument("-srs", default=None, type=str)
raster_parser.add_argument("-place", default=None, type=str)
raster_parser.add_argument("-outfile_basename", default=None, type=str)


def handle(args):
    if args.group == "auth":
        auth_handler(args)
    elif args.group == "metadata":
        metadata_handler(args)
    elif args.group == "places":
        places_handler(args)
    elif args.group == "raster":
        raster_handler(args)
    else:
        print("invalid command group")


if __name__ == "__main__":
    handle(parser.parse_args())



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import mock
import os
from six import StringIO
from descarteslabs.scripts.parser import parser, handle
import base64
import json

token = base64.b64encode(
    json.dumps({
        "client_id": os.environ.get("CLIENT_ID"),
        "client_secret": os.environ.get("CLIENT_SECRET")}).encode(
        'utf-8'))


class TestScripts(unittest.TestCase):
    def test_auth_login(self):
        with mock.patch('descarteslabs.scripts.parser.auth.input', return_value=token):
            handle(parser.parse_args(["auth", "login"]))

        with mock.patch('sys.stdout', new_callable=StringIO) as out:
            handle(parser.parse_args(["auth", "groups"]))
            self.assertEqual(out.getvalue().strip(), '["public"]')

    def test_places_find(self):
        with mock.patch('sys.stdout', new_callable=StringIO) as out:
            handle(parser.parse_args(["places", "find", "iowa"]))
            iowa = [
                {
                    "name": "Iowa",
                    "bbox": [
                        - 96.639468,
                        40.37544,
                        -90.140061,
                        43.501128
                    ],
                    "id": 85688713,
                    "path": "continent:north-america_country:united-states_region:iowa",
                    "slug": "north-america_united-states_iowa",
                    "placetype": "region"
                }
            ]
            self.assertEqual(json.loads(out.getvalue().strip()), iowa)

    def test_metadata_sources(self):
        with mock.patch('sys.stdout', new_callable=StringIO) as out:
            handle(parser.parse_args(["metadata", "sources"]))
            sources = [{'sat_id': 'LANDSAT_8', 'product': 'landsat:LC08:PRE:TOAR'}]
            self.assertEqual(json.loads(out.getvalue().strip()), sources)


if __name__ == '__main__':
    unittest.main()



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import unittest
from warnings import catch_warnings

import descarteslabs as dl
from descarteslabs.exceptions import NotFoundError


class TestMetadata(unittest.TestCase):
    instance = None

    @classmethod
    def setUpClass(cls):
        cls.instance = dl.metadata

    def test_sources(self):
        r = self.instance.sources()
        self.assertGreater(len(r), 0)

    def test_available_products(self):
        r = self.instance.available_products()
        self.assertGreater(len(r), 0)

    def test_search(self):
        r = self.instance.search()
        self.assertGreater(len(r['features']), 0)

    def test_search_dltile(self):
        dltile = "256:16:30.0:15:-11:591"
        r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07',
                                 const_id=['L8'], dltile=dltile)
        keys = [f['key'] for f in r['features']]
        self.assertTrue('meta_LC80270312016188_v1' in keys)

    def test_sat_id(self):
        r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', sat_id='LANDSAT_8')
        self.assertGreater(len(r['features']), 0)

    def test_cloud_fraction(self):
        r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', sat_id='LANDSAT_8',
                                 cloud_fraction=0.5)
        for feature in r['features']:
            self.assertLess(feature['properties']['cloud_fraction'], 0.5)

        r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', sat_id='LANDSAT_8',
                                 cloud_fraction=0.0)
        for feature in r['features']:
            self.assertEqual(feature['properties']['cloud_fraction'], 0.0)

    def test_search_by_product(self):
        r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', products=['landsat:LC08:PRE:TOAR'])
        self.assertGreater(len(r['features']), 0)

    def test_fields(self):
        cases = [
            [],
            ["id", "key"],  # ["id", "key"]
            ["key"],  # ["id", "key"]
            ["geometry"],  # ["geometry", "id", "type"]
        ]

        for fields in cases:
            r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', fields=fields)
            for feature in r['features']:
                if "id" not in fields:
                    fields.append("id")

                if "geometry" in fields:
                    fields.append("type")

                if "key" in fields or "geometry" in fields:
                    fields.append("properties")

                self.assertEqual(sorted(feature.keys()), sorted(fields))

        fields = ["cloud_fraction", "fill_fraction"]
        r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', fields=fields)
        for feature in r['features']:
            self.assertEqual(sorted(feature['properties'].keys()), sorted(fields))

    def test_const_id_search_deprecation(self):
        with catch_warnings(record=True) as w:
            r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', const_id=['l8'])
            self.assertGreater(len(r['features']), 0)
            self.assertEqual(len(w), 1)
            self.assertIn('deprecated', str(w[0].message))

    def test_multiple_products_search(self):
        r = self.instance.search(
            start_time='2016-07-06',
            end_time='2016-07-07',
            products=['landsat:LE07:PRE:TOAR', 'landsat:LC08:PRE:TOAR']
        )
        self.assertGreater(len(r['features']), 0)

    def test_place(self):
        r = self.instance.search(products=['landsat:LC08:PRE:TOAR'], place='north-america_united-states_iowa', limit=1)
        self.assertEqual(1, len(r['features']))

    def test_summary(self):
        r = self.instance.summary(
            start_time='2016-07-06',
            end_time='2016-07-07',
            products=['landsat:LC08:PRE:TOAR'],
            pixels=True
        )
        self.assertIn('products', r)
        self.assertIn('count', r)
        self.assertIn('pixels', r)
        self.assertIn('bytes', r)
        self.assertGreater(r['count'], 0)

    def test_const_id_summary_deprecation(self):
        with catch_warnings(record=True) as w:
            r = self.instance.summary(
                start_time='2016-07-06',
                end_time='2016-07-07',
                const_id=['l8'],
                pixels=True
            )
            self.assertEqual(len(w), 1)
            self.assertIn('deprecated', str(w[0].message))

            self.assertIn('products', r)
            self.assertIn('count', r)
            self.assertIn('pixels', r)
            self.assertIn('bytes', r)
            self.assertGreater(r['count'], 0)

    def test_summary_part(self):
        r = self.instance.summary(
            start_time='2016-07-06',
            end_time='2016-07-07',
            products=['landsat:LC08:PRE:TOAR'],
            part='year',
            pixels=True
        )
        self.assertIn('count', r)
        self.assertIn('pixels', r)
        self.assertIn('bytes', r)
        self.assertIn('items', r)
        self.assertEqual(len(r['items']), 1)

    def test_features(self):
        r = self.instance.features(start_time='2016-07-06', end_time='2016-07-07', sat_id='LANDSAT_8', limit=20)
        first_21 = itertools.islice(r, 21)
        self.assertGreater(len(list(first_21)), 0)

    def test_products_search_sort(self):
        r = self.instance.search(start_time='2016-07-06', end_time='2016-07-07', products=['landsat:LC08:PRE:TOAR'],
                                 sort_field="cloud_fraction", sort_order="desc")
        self.assertGreater(len(r['features']), 0)

        last = None
        for feature in r['features']:
            current = feature['properties']['cloud_fraction']
            if last is None:
                last = current
                continue

            self.assertLessEqual(current, last)

    def test_search_products(self):
        r = self.instance.products(limit=1)
        self.assertEqual(len(r), 1)

    def test_search_bands(self):
        r = self.instance.bands()
        # no bands are publicly available.
        self.assertEqual(len(r), 0)

    def test_products_get(self):
        product_id = 'landsat:LC08:PRE:TOAR'
        r = self.instance.get_product(product_id)
        self.assertEqual(r['product'], product_id)

    def test_bands_get(self):
        band_id = 'landsat:LC08:PRE:TOAR:red'
        try:
            self.instance.get_product(band_id)
        except NotFoundError:
            pass


if __name__ == '__main__':
    unittest.main()



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


def as_json_string(str_or_dict):
    if not str_or_dict:
        return str_or_dict
    elif isinstance(str_or_dict, dict):
        return json.dumps(str_or_dict)
    else:
        return str_or_dict



class ClientError(Exception):
    """ Base class """
    pass


class AuthError(ClientError):
    pass


class OauthError(AuthError):
    pass


class ServerError(Exception):
    status = 500


class BadRequestError(ServerError):
    status = 400


class NotFoundError(ServerError):
    status = 404


class RateLimitError(ServerError):
    status = 429


class GatewayTimeoutError(ServerError):
    status = 504



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from descarteslabs.auth import Auth
import requests
import json


class TestAuth(unittest.TestCase):
    def test_get_token(self):
        # get a jwt
        auth = Auth()
        self.assertIsNotNone(auth.token)

        # validate the jwt
        url = auth.domain + "/tokeninfo"
        params = {"id_token": auth.token}
        headers = {"content-type": "application/json"}
        r = requests.post(url, data=json.dumps(params), headers=headers)
        self.assertEqual(200, r.status_code)


if __name__ == '__main__':
    unittest.main()



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import os
from io import BytesIO
import json
import warnings

from descarteslabs.addons import numpy as np
from descarteslabs.utilities import as_json_string
from .service import Service
from .places import Places
import six


class Raster(Service):
    """Raster"""
    TIMEOUT = (9.5, 300)

    def __init__(self, url=None, token=None):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        if url is None:
            url = os.environ.get("DESCARTESLABS_RASTER_URL", "https://platform-services.descarteslabs.com/raster/v1")

        Service.__init__(self, url, token)

    def get_bands_by_key(self, key):
        """
        For a given source id, return the available bands.

        :param str key: A :class:`Metadata` identifiers.

        :return: A dictionary of band entries and their metadata.
        """
        r = self.session.get('/bands/key/%s' % key)

        return r.json()

    def get_bands_by_constellation(self, const):
        """
        For a given constellation id, return the available bands.

        :param str const: A constellation name/abbreviation.

        :return: A dictionary of band entries and their metadata.
        """
        r = self.session.get('/bands/constellation/%s' % const)
        return r.json()

    def dltiles_from_shape(self, resolution, tilesize, pad, shape):
        """
        Return a feature collection of DLTile GeoJSONs that intersect
        a GeoJSON Geometry `shape`.

        :param float resolution: Resolution of DLTile
        :param int tilesize: Number of valid pixels per DLTile
        :param int pad: Number of ghost pixels per DLTile (overlap among tiles)
        :param str shape: A GeoJSON geometry specifying a shape over
            which to intersect DLTiles.

        :return: GeoJSON FeatureCollection of intersecting DLTile geometries.

        Example::

            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> iowa = dl.places.shape("north-america_united-states_iowa")
            >>> tiles = dl.raster.dltiles_from_shape(30.0, 2048, 16, iowa)
            >>> pprint(tiles['features'][0])
            {'geometry': {'coordinates': [[[-96.81264975325391, 41.04520331997488],
                                           [-96.07101667769165, 41.02873098011615],
                                           [-96.04576296033328, 41.590072611375206],
                                           [-96.79377566762062, 41.606871549447824],
                                           [-96.81264975325391, 41.04520331997488]]],
                          'type': 'Polygon'},
             'properties': {'cs_code': 'EPSG:32614',
                            'key': '2048:16:30.0:14:3:74',
                            'outputBounds': [683840.0, 4546080.0, 746240.0, 4608480.0],
                            'pad': 16,
                            'resolution': 30.0,
                            'ti': 3,
                            'tilesize': 2048,
                            'tj': 74,
                            'zone': 14},
             'type': 'Feature'}

        """

        shape = as_json_string(shape)
        params = {
            'resolution': resolution,
            'tilesize': tilesize,
            'pad': pad,
            'shape': shape,
        }

        r = self.session.post('/dlkeys/from_shape', json=params)
        return r.json()

    def dltile_from_latlon(self, lat, lon, resolution, tilesize, pad):
        """
        Return a DLTile GeoJSON Feature that covers a latitude/longitude

        :param float lat: Requested latitude
        :param float lon: Requested longitude
        :param float resolution: Resolution of DLTile
        :param int tilesize: Number of valid pixels per DLTile
        :param int pad: Number of ghost pixels per DLTile (overlap among tiles)

        :return: A DLTile GeoJSON Feature

        Example::

            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> pprint(dl.raster.dltile_from_latlon(45, 60, 15.0, 1024, 16))
            {'geometry': {'coordinates': [[[59.88428127486957, 44.89851158838881],
                                           [60.084634558186266, 44.903806716073376],
                                           [60.07740397456606, 45.04621255053833],
                                           [59.87655568676388, 45.04089121582091],
                                           [59.88428127486957, 44.89851158838881]]],
                          'type': 'Polygon'},
            'properties': {'cs_code': 'EPSG:32641',
                           'key': '1024:16:15.0:41:-16:324',
                           'outputBounds': [254000.0, 4976400.0, 269840.0, 4992240.0],
                           'pad': 16,
                           'resolution': 15.0,
                           'ti': -16,
                           'tilesize': 1024,
                           'tj': 324,
                           'zone': 41},
            'type': 'Feature'}
        """
        params = {
            'resolution': resolution,
            'tilesize': tilesize,
            'pad': pad,
        }

        r = self.session.get('/dlkeys/from_latlon/%f/%f' % (lat, lon), params=params)

        return r.json()

    def dltile(self, key):
        """
        Given a DLTile key, return a DLTile GeoJSON Feature

        :param str key: A DLTile key that identifies a DLTile

        :return: A DLTile GeoJSON Feature

        Example::

            >>> import descarteslabs as dl
            >>> from pprint import pprint
            >>> pprint(dl.raster.dltile("1024:16:15.0:41:-16:324"))
            {'geometry': {'coordinates': [[[59.88428127486957, 44.89851158838881],
                                           [60.084634558186266, 44.903806716073376],
                                           [60.07740397456606, 45.04621255053833],
                                           [59.87655568676388, 45.04089121582091],
                                           [59.88428127486957, 44.89851158838881]]],
                          'type': 'Polygon'},
             'properties': {'cs_code': 'EPSG:32641',
                            'key': '1024:16:15.0:41:-16:324',
                            'outputBounds': [254000.0, 4976400.0, 269840.0, 4992240.0],
                            'pad': 16,
                            'resolution': 15.0,
                            'ti': -16,
                            'tilesize': 1024,
                            'tj': 324,
                            'zone': 41},
             'type': 'Feature'}
        """

        r = self.session.get('/dlkeys/%s' % key)

        return r.json()

    def dlkeys_from_shape(self, resolution, tilesize, pad, shape):
        """
        Deprecated. See dltiles_from_shape
        """
        warnings.warn("dlkey methods have been renamed to dltile",
                      DeprecationWarning, stacklevel=2)
        return self.dltiles_from_shape(resolution, tilesize, pad, shape)

    def dlkey_from_latlon(self, lat, lon, resolution, tilesize, pad):
        """
        Deprecated. See dltile_from_latlon
        """
        warnings.warn("dlkey methods have been renamed to dltile",
                      DeprecationWarning, stacklevel=2)
        return self.dltile_from_latlon(lat, lon, resolution, tilesize, pad)

    def dlkey(self, key):
        """
        Deprecated. See dltile
        """
        warnings.warn("dlkey methods have been renamed to dltile",
                      DeprecationWarning, stacklevel=2)
        return self.dltile(key)

    def raster(
            self,
            inputs,
            bands=None,
            scales=None,
            data_type=None,
            output_format='GTiff',
            srs=None,
            dimensions=None,
            resolution=None,
            bounds=None,
            bounds_srs=None,
            cutline=None,
            place=None,
            align_pixels=False,
            resampler=None,
            dltile=None,
            save=False,
            outfile_basename=None,
    ):
        """Given a list of :class:`Metadata <descarteslabs.services.Metadata>` identifiers,
        retrieve a translated and warped mosaic.

        :param inputs: List of :class:`Metadata` identifiers.
        :param bands: List of requested bands.
        :param scales: List of tuples specifying the scaling to be applied to each band.
            If no scaling is desired for a band, use ``None`` where appropriate. If a
            tuple contains four elements, the last two will be used as the output range.
            For example, ``(0, 10000, 0, 128)`` would scale the source values 0-10000 to
            be returned as 0-128 in the output.
        :param str output_format: Output format (`GTiff`, `PNG`, ...).
        :param str data_type: Output data type (`Byte`, `UInt8`, `UInt16`, `Float32`, etc).
        :param str srs: Output spatial reference system definition understood by GDAL.
        :param float resolution: Desired resolution in output SRS units. Incompatible with
            `dimensions`
        :param tuple dimensions: Desired output (width, height) in pixels. Incompatible with
            `resolution`
        :param str cutline: A GeoJSON feature or geometry to be used as a cutline.
        :param str place: A slug identifier to be used as a cutline.
        :param tuple bounds: ``(min_x, min_y, max_x, max_y)`` in target SRS.
        :param str bounds_srs: Override the coordinate system in which bounds are expressed.
        :param bool align_pixels: Align pixels to the target coordinate system.
        :param str resampler: Resampling algorithm to be used during warping (``near``,
            ``bilinear``, ``cubic``, ``cubicsplice``, ``lanczos``, ``average``, ``mode``,
            ``max``, ``min``, ``med``, ``q1``, ``q3``).
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param bool save: Write resulting files to disk. Default: False
        :param str outfile_basename: If 'save' is True, override default filename using
            this string as a base.
        """
        cutline = as_json_string(cutline)

        if place:
            places = Places()
            places.auth = self.auth
            shape = places.shape(place, geom='low')
            cutline = json.dumps(shape['geometry'])

        params = {
            'keys': inputs,
            'bands': bands,
            'scales': scales,
            'ot': data_type,
            'of': output_format,
            'srs': srs,
            'resolution': resolution,
            'shape': cutline,
            'outputBounds': bounds,
            'outputBoundsSRS': bounds_srs,
            'outsize': dimensions,
            'targetAlignedPixels': align_pixels,
            'resampleAlg': resampler,
        }

        if dltile is not None:
            if isinstance(dltile, dict):
                params['dltile'] = dltile['properties']['key']
            else:
                params['dltile'] = dltile

        r = self.session.post('/raster', json=params)

        json_resp = r.json()
        # Decode base64
        for k in list(json_resp['files'].keys()):
            if outfile_basename:
                outfilename = "{}.{}".format(
                    outfile_basename,
                    ".".join(os.path.basename(k).split(".")[1:])
                )
            else:
                outfilename = k
            json_resp['files'][outfilename] = base64.b64decode(
                json_resp['files'].pop(k)
            )

        if save:
            for filename, data in six.iteritems(json_resp['files']):
                with open(filename, "wb") as f:
                    f.write(data)

        return json_resp

    def ndarray(
            self,
            inputs,
            bands=None,
            scales=None,
            data_type=None,
            srs=None,
            resolution=None,
            dimensions=None,
            cutline=None,
            place=None,
            bounds=None,
            bounds_srs=None,
            align_pixels=False,
            resampler=None,
            order='image',
            dltile=None,
    ):
        """Retrieve a raster as a NumPy array.

        :param inputs: List of :class:`Metadata` identifiers.
        :param bands: List of requested bands.
        :param scales: List of tuples specifying the scaling to be applied to each band.
            If no scaling is desired for a band, use ``None`` where appropriate. If a
            tuple contains four elements, the last two will be used as the output range.
            For example, ``(0, 10000, 0, 128)`` would scale the source values 0-10000 to
            be returned as 0-128 in the output.
        :param str data_type: Output data type (`Byte`, `UInt8`, `UInt16`, `Float32`, etc).
        :param str srs: Output spatial reference system definition understood by GDAL.
        :param float resolution: Desired resolution in output SRS units. Incompatible with
            `dimensions`
        :param tuple dimensions: Desired output (width, height) in pixels. Incompatible with
            `resolution`
        :param str cutline: A GeoJSON feature or geometry to be used as a cutline.
        :param str place: A slug identifier to be used as a cutline.
        :param tuple bounds: ``(min_x, min_y, max_x, max_y)`` in target SRS.
        :param str bounds_srs: Override the coordinate system in which bounds are expressed.
        :param bool align_pixels: Align pixels to the target coordinate system.
        :param str resampler: Resampling algorithm to be used during warping (``near``,
            ``bilinear``, ``cubic``, ``cubicsplice``, ``lanczos``, ``average``, ``mode``,
            ``max``, ``min``, ``med``, ``q1``, ``q3``).
        :param str order: Order of the returned array. `image` returns arrays as
            ``(row, column, band)`` while `gdal` returns arrays as ``(band, row, column)``.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.

        """
        cutline = as_json_string(cutline)

        if place is not None:
            places = Places()
            places.auth = self.auth
            shape = places.shape(place, geom='low')
            cutline = json.dumps(shape['geometry'])

        params = {
            'keys': inputs,
            'bands': bands,
            'scales': scales,
            'ot': data_type,
            'srs': srs,
            'resolution': resolution,
            'shape': cutline,
            'outputBounds': bounds,
            'outputBoundsSRS': bounds_srs,
            'outsize': dimensions,
            'targetAlignedPixels': align_pixels,
            'resampleAlg': resampler,
        }

        if dltile is not None:
            if isinstance(dltile, dict):
                params['dltile'] = dltile['properties']['key']
            else:
                params['dltile'] = dltile

        r = self.session.post('/npz', json=params)

        io = BytesIO(r.content)
        npz = np.load(io)
        array = npz['data']
        metadata = json.loads(npz['metadata'].tostring().decode('utf-8'))

        if len(array.shape) > 2:
            if order == 'image':
                return array.transpose((1, 2, 0)), metadata
            elif order == 'gdal':
                return array, metadata
        else:
            return array, metadata



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

import random
import base64
import json
import datetime
import six
import os
import stat

from descarteslabs.exceptions import AuthError, OauthError


def base64url_decode(input):
    """Helper method to base64url_decode a string.
    Args:
        input (str): A base64url_encoded string to decode.
    """
    rem = len(input) % 4
    if rem > 0:
        input += b'=' * (4 - rem)

    return base64.urlsafe_b64decode(input)


class Auth:
    def __init__(self, domain="https://descarteslabs.auth0.com",
                 scope=None, leeway=500):
        """
        Helps retrieve JWT from a client id and refresh token for cli usage.
        :param client_id: str
        :param refresh_token: str generated through IAM interface
        :param url: endpoint for auth0
        :param scope: the JWT fields to be included
        :param leeway: JWT expiration leeway
        """

        token_info = {}

        try:
            with open(os.path.join(os.path.expanduser("~"), '.descarteslabs', 'token_info.json')) as fp:
                token_info = json.load(fp)
        except:
            pass

        self.client_id = os.environ.get('CLIENT_ID', token_info.get('client_id', None))
        self.client_secret = os.environ.get('CLIENT_SECRET', token_info.get('client_secret', None))
        self._token = os.environ.get('JWT_TOKEN', token_info.get('jwt_token', None))

        self.domain = domain
        self.scope = scope
        self.leeway = leeway

        if self.scope is None:
            self.scope = ['openid', 'name', 'groups']

    @property
    def token(self):
        if self._token is None:
            self._get_token()

        exp = self.payload.get('exp')

        if exp is not None:
            now = (datetime.datetime.utcnow() - datetime.datetime(1970, 1, 1)).total_seconds()
            if now + self.leeway > exp:
                self._get_token()

        return self._token

    @property
    def payload(self):
        if self._token is None:
            self._get_token()

        if isinstance(self._token, six.text_type):
            token = self._token.encode('utf-8')
        else:
            token = self._token

        claims = token.split(b'.')[1]
        return json.loads(base64url_decode(claims).decode('utf-8'))

    def _get_token(self, timeout=100):
        if self.client_id is None:
            raise AuthError("Could not find CLIENT_ID")

        if self.client_secret is None:
            raise AuthError("Could not find CLIENT_SECRET")

        s = requests.Session()
        retries = Retry(total=5,
                        backoff_factor=random.uniform(1, 10),
                        method_whitelist=frozenset(['GET', 'POST']),
                        status_forcelist=[429, 500, 502, 503, 504])

        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"content-type": "application/json"}
        params = {
            "scope": " ".join(self.scope),
            "client_id": self.client_id,
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "target": self.client_id,
            "api_type": "app",
            "refresh_token": self.client_secret
        }
        r = s.post(self.domain + "/delegation", headers=headers, data=json.dumps(params), timeout=timeout)

        if r.status_code != 200:
            raise OauthError("%s: %s" % (r.status_code, r.text))

        data = r.json()

        self._token = data['id_token']

        token_info = {}

        try:
            with open(os.path.join(os.path.expanduser("~"), '.descarteslabs', 'token_info.json')) as fp:
                token_info = json.load(fp)
        except:
            pass

        token_info['jwt_token'] = self._token

        path = os.path.join(os.path.expanduser("~"), '.descarteslabs')

        if not os.path.exists(path):
            os.makedirs(path)

        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

        file = os.path.join(os.path.expanduser("~"), '.descarteslabs', 'token_info.json')

        with open(file, 'w+') as fp:
            json.dump(token_info, fp)

        os.chmod(file, stat.S_IRUSR | stat.S_IWUSR)


if __name__ == '__main__':
    auth = Auth()

    print(auth.token)



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

import descarteslabs
from ..exceptions import ServerError, BadRequestError, NotFoundError, RateLimitError, GatewayTimeoutError


class WrappedSession(requests.Session):
    def __init__(self, base_url, timeout=None):
        self.base_url = base_url
        self.timeout = timeout
        super(WrappedSession, self).__init__()

    def request(self, method, url, **kwargs):
        if self.timeout and 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout

        resp = super(WrappedSession, self).request(method, self.base_url + url, **kwargs)

        if resp.status_code == 200:
            return resp
        elif resp.status_code == 400:
            raise BadRequestError(resp.text)
        elif resp.status_code == 404:
            raise NotFoundError("404 %s %s" % (method, url))
        elif resp.status_code == 429:
            raise RateLimitError(resp.text)
        elif resp.status_code == 504:
            raise GatewayTimeoutError(
                "Your request timed out on the server. "
                "Consider reducing the complexity of your request.")
        else:
            raise ServerError(resp.text)


class Service:
    TIMEOUT = (9.5, 30)

    def __init__(self, url, token):
        self.auth = descarteslabs.descartes_auth
        self.base_url = url
        if token:
            self.auth._token = token

        self.current_session = None
        self.session_token = None

    @property
    def token(self):
        return self.auth.token

    @token.setter
    def token(self, token):
        self.auth._token = token

    @property
    def session(self):
        if self.session_token != self.token:
            self.current_session = None

        if not self.current_session:
            self.current_session = self.build_session()
            self.session_token = self.token

        return self.current_session

    def build_session(self):
        s = WrappedSession(self.base_url, timeout=self.TIMEOUT)

        retries = Retry(total=5,
                        read=2,
                        backoff_factor=random.uniform(1, 3),
                        method_whitelist=frozenset([
                            'HEAD', 'TRACE', 'GET', 'POST',
                            'PUT', 'OPTIONS', 'DELETE'
                        ]),
                        status_forcelist=[429, 500, 502, 503])

        s.mount('https://', HTTPAdapter(max_retries=retries))

        s.headers.update({
            "Authorization": self.token,
            "Content-Type": "application/json",
            "User-Agent": "dl-python/{}".format(descarteslabs.__version__)
        })

        here = os.path.dirname(__file__)

        try:
            file = os.path.join(here, 'gd_bundle-g2-g1.crt')
            with open(file):
                s.verify = file
        except:
            s.verify = False

        return s



# Copyright 2017 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import descarteslabs as dl


class TestPlaces(unittest.TestCase):
    instance = None

    @classmethod
    def setUpClass(cls):
        cls.instance = dl.places

    def test_placetypes(self):
        data = self.instance.placetypes()
        self.assertEqual(10, len(data))

    def test_find(self):
        r = self.instance.find('united-states_iowa')
        self.assertEqual(1, len(r))

    def test_shape(self):
        r = self.instance.shape('north-america_united-states_iowa')
        self.assertEqual(85688713, r['id'])

    def test_prefix(self):
        # counties by default
        r = self.instance.prefix('north-america_united-states_iowa')
        self.assertEqual(99, len(r['features']))

        r = self.instance.prefix('north-america_united-states_iowa', placetype='district')
        self.assertEqual(9, len(r['features']))


if __name__ == '__main__':
    unittest.main()
